package org.bitbucket.bradleysmithllc.etlunit.feature.file;

import com.google.inject.Binder;
import com.google.inject.Injector;
import com.google.inject.Module;
import org.bitbucket.bradleysmithllc.etlunit.*;
import org.bitbucket.bradleysmithllc.etlunit.context.VariableContext;
import org.bitbucket.bradleysmithllc.etlunit.feature.AbstractFeature;
import org.bitbucket.bradleysmithllc.etlunit.feature.FeatureModule;
import org.bitbucket.bradleysmithllc.etlunit.feature.assertion.AssertionFeatureModule;
import org.bitbucket.bradleysmithllc.etlunit.feature.file.json.file._assert.AssertHandler;
import org.bitbucket.bradleysmithllc.etlunit.feature.file.json.file._assert.AssertRequest;
import org.bitbucket.bradleysmithllc.etlunit.feature.file.json.file.stage.StageHandler;
import org.bitbucket.bradleysmithllc.etlunit.feature.file.json.file.stage.StageRequest;
import org.bitbucket.bradleysmithllc.etlunit.parser.*;
import org.bitbucket.bradleysmithllc.etlunit.util.JSonBuilderProxy;

import javax.inject.Inject;
import java.io.*;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

/**
 * Staged file objects look like this:
 * fileContext_[contextName] [
 * {
 * file: '/path/to/file'
 * (, classifier: 'variableName')?
 * (, destination-name: 'destinationName')?
 * }(,{ . . . })*
 * ]
 * <p/>
 * If this is the default context, then the variable name will be fileContext_DEFAULT.
 * Where file is the path to the file to be utilized, and classifier is a meaning that is specific to the file consumer.
 */

@FeatureModule
public class FileFeatureModule extends AbstractFeature
{
	FileAssertExtender fileAssertExtender = new FileAssertExtender(this);

	private static final List<String> prerequisites = Arrays.asList("logging", "assertion");

	private static final String CONTEXT_PREFIX = "fileContext[";
	private static final String CONTEXT_POSTFIX = "]";
	private static final String DEFAULT_CONTEXT_NAME = "DEFAULT";

	private AssertionFeatureModule assertionFeatureModule;
	private final FileRuntimeSupport fileRuntimeSupport = new FileRuntimeSupportImpl();

	private final class RequestHandler extends NullClassListener implements StageHandler,
			AssertHandler
	{
		public action_code stage(StageRequest operation, ETLTestMethod mt, ETLTestOperation op, ETLTestValueObject obj, VariableContext context, ExecutionContext econtext) throws TestAssertionFailure, TestExecutionError, TestWarning
		{
			String file = operation.getFile();
			String listFile = operation.getListFile();

			if (file == null)
			{
				if (listFile == null)
				{
					throw new IllegalArgumentException("Either one of list-file or file must be specified");
				}

				file = listFile;
			}

			List<String> fileNames = new ArrayList<String>();
			fileNames.add(file);

			String package_ = op.getTestMethod().getTestClass().getPackage();

			if (listFile != null)
			{
				// blast all names contained in the list file into the context
				File source = fileRuntimeSupport.getDataFile(package_, listFile);

				if (!source.exists())
				{
					throw new TestExecutionError(listFile, FileConstants.ERR_SOURCE_FILE_MISSING);
				}

				try
				{
					BufferedReader bufferedReader = new BufferedReader(new FileReader(source));

					try
					{
						String str = null;

						while ((str = bufferedReader.readLine()) != null)
						{
							fileNames.add(str);
						}
					}
					finally
					{
						bufferedReader.close();
					}
				}
				catch (IOException e)
				{
					throw new TestExecutionError("List file does not exist", e);
				}
			}

			String var = operation.getVariableName();

			for (int i = 0; i < fileNames.size(); i++)
			{
				String fileName = fileNames.get(i);

				File source = fileRuntimeSupport.getDataFile(package_, fileName);

				if (!source.exists())
				{
					throw new TestExecutionError(fileName, FileConstants.ERR_SOURCE_FILE_MISSING);
				}

				String classifier = operation.getClassifier();

				JSonBuilderProxy proxy = new JSonBuilderProxy().object().key("file").value(source.getAbsolutePath());

				if (classifier != null)
				{
					proxy = proxy.key("classifier").value(classifier);
				}

				String destinationName = operation.getDestinationName();

				if (destinationName != null)
				{
					proxy = proxy.key("destination-name").value(destinationName);
				}

				// Set the file variable.  For list files this will be @var for the first, and @var-2..n for the second on (1-based numbering)
				if (var != null)
				{
					context.declareAndSetStringValue(var + (i == 0 ? "" : ("-" + (i + 1))), destinationName != null ? destinationName : fileName);
				}

				String str = proxy.endObject().toString();

				ETLTestValueObject co = getFileContext(obj, context);

				if (co != null)
				{
					// this will be an array of objects.  We need to add another to the end
					try
					{
						co.getValueAsList().add(ETLTestParser.loadObject(str));
					}
					catch (ParseException e)
					{
						throw new IllegalArgumentException(e);
					}
				}
				else
				{
					setFileContext(obj, context, "[" + str + "]");
				}
			}

			return action_code.handled;
		}

		@Override
		public action_code _assert(AssertRequest request, ETLTestMethod mt, ETLTestOperation testOperation, ETLTestValueObject valueObject, VariableContext variableContext, ExecutionContext executionContext)
				throws TestAssertionFailure, TestExecutionError, TestWarning
		{
			return fileAssertExtender._assert(request, mt, testOperation, valueObject, variableContext, executionContext);
		}
	}

	private final RequestHandler stageHandler = new RequestHandler();

	@Override
	public void initialize(Injector inj)
	{
		postCreate(fileAssertExtender);

		//assertionFeatureModule.extend();
	}

	@Override
	protected Injector preCreateSub(Injector inj)
	{
		return inj.createChildInjector(new Module()
		{
			public void configure(Binder binder)
			{
				binder.bind(FileRuntimeSupport.class).toInstance(fileRuntimeSupport);
			}
		});
	}

	@Inject
	public void setAssertionFeature(AssertionFeatureModule f)
	{
		assertionFeatureModule = f;
	}

	@Override
	public List<String> getPrerequisites()
	{
		return prerequisites;
	}

	@Override
	protected List<String> getSupportedFolderNamesSub()
	{
		return Arrays.asList("files");
	}

	public String getFeatureName()
	{
		return "file";
	}

	@Override
	public ClassListener getListener()
	{
		return stageHandler;
	}

	private static String getFileContextId(ETLTestValueObject param)
	{
		ETLTestValueObject contextNameObj = param.query("context-name");

		String contextName = null;

		if (contextNameObj != null)
		{
			contextName = contextNameObj.getValueAsString();
		}

		return getContextKey(contextName);
	}

	private static void setFileContext(ETLTestValueObject param, VariableContext vcontext, String json)
	{
		String fileContextId = getFileContextId(param);

		if (!vcontext.hasVariableBeenDeclared(fileContextId))
		{
			try
			{
				vcontext.declareAndSetValue(fileContextId, ETLTestParser.loadObject("{}"));
			}
			catch (ParseException e)
			{
				throw new RuntimeException(e);
			}
		}

		ETLTestValueObject fcontext = getFileContextContainer(param, vcontext);

		new ETLTestValueObjectBuilder(fcontext).key("file-list").jsonValue(json);
	}

	public static ETLTestValueObject getFileContext(ETLTestValueObject param, VariableContext vcontext)
	{
		ETLTestValueObject context = getFileContextContainer(param, vcontext);

		if (context != null)
		{
			return context.query("file-list");
		}

		return null;
	}

	public static ETLTestValueObject getFileContextContainer(ETLTestValueObject param, VariableContext vcontext)
	{
		String contextKey = getFileContextId(param);

		if (vcontext.hasVariableBeenDeclared(contextKey))
		{
			return vcontext.getValue(contextKey);
		}

		return null;
	}

	/**
	 * @param id - null for default context
	 * @return
	 */
	private static String getContextKey(String id)
	{
		return CONTEXT_PREFIX + (id == null ? DEFAULT_CONTEXT_NAME : id) + CONTEXT_POSTFIX;
	}
}